﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ElfViewerConsole.ELFModel.ELFFormat;
using System.IO;

namespace ElfViewerConsole
{
    class ConsoleOutput
    {
        private ELFFile ef;
        private static readonly string FormatString_Default = "{0,-16}{1,-18}{2,-18}{3,-18}{4,-28}";
        private static readonly string FormatString_ByteArray = "{0,-16}{1,8}    ";
        private static readonly string FormatString_Default_Short = "{0,-16}{1,22}";

        private static readonly string FormatString_SecTab_Short = "{0,-8}{1,-18}{2,-18}{3,-8}{4,-18}{5,-18}{6,-18}{7,-12}{8,-12}{9,-18}{10,-12}";
        private static readonly string FormatString_SymTab_Short = "{0,-8}{1,-18}{2,-18}{3,-12}{4,-12}{5,-12}{6,-12}{7,-40}";
        private static readonly string FormatString_ProgTab_Short = "{0,-12}{1,-18}{2,-18}{3,-18}{4,-18}{5,-18}{6,-18}{7,-12}";

        public ConsoleOutput(ELFFile elffile)
        {
            this.ef = elffile;
        }

        public void PrintGeneralFileInfo()
        {
            const string formatStr = "{0, -15}{1, -80}";

            Console.WriteLine("General file info:");
            Console.WriteLine("------------------");
            Console.WriteLine(formatStr, "File name:", ef.FileName);
            Console.WriteLine(formatStr, "Full path:", ef.FilePath);
            Console.WriteLine(formatStr, "Size (bytes):", ef.FileSize);
            Console.WriteLine(formatStr, "Creation:", String.Format("{0:G}", ef.FileInfo.CreationTime));
            Console.WriteLine(formatStr, "Last access:", String.Format("{0:G}", ef.FileInfo.LastAccessTime));
            Console.WriteLine(formatStr, "Last write:", String.Format("{0:G}", ef.FileInfo.LastWriteTime));
            Console.WriteLine(formatStr, "MD5 Hash", FileUtil.GetMD5Hash(ef));
            Console.WriteLine(formatStr, "SHA-1 Hash", FileUtil.GetSHA1Hash(ef));
            Console.WriteLine(formatStr, "SHA-256 Hash", FileUtil.GetSHA256Hash(ef));
        }

        public void PrintElfHeaderDetailed()
        {
            Console.Out.WriteLine();
            Console.Out.WriteLine("ELF header:");
            Console.Out.WriteLine("-----------");
            PrintLine("Name", "Offset", "NumValue", "Value", "Description");

            foreach (IELFElem elem in ef.ElfFileHeader)
            {
                if (elem.IsEnumValue())
                {
                    PrintLineWithEnum(elem);
                }
                else
                {
                    PrintLine(elem);
                }
            }
        }

        public void PrintElfHeaderShort()
        {
            Console.Out.WriteLine();
            Console.Out.WriteLine("ELF header:");
            Console.Out.WriteLine("-----------");
            foreach (IELFElem elem in ef.ElfFileHeader)
            {
                if (elem.IsEnumValue())
                {
                    PrintLineWithEnumShort(elem);
                }
                else
                {
                    PrintLineShort(elem);
                }
            }
        }

        public void PrintProgramHeaderTableDetailed()
        {
            Console.Out.WriteLine();
            Console.Out.WriteLine("ELF Program Header Tables:");
            Console.Out.WriteLine("--------------------------");

            if (!ef.ExistsProgramHeaderTable())
            {
                Console.Out.WriteLine("Program header tables do not exist.");
                return;
            }

            foreach (ELFProgramHeaderTable table in ef.GetProgramHeaderTables())
            {
                Console.WriteLine("{0,10}", "Prorgram header table " + table.HeaderTableIndex);

                PrintLine("Name", "Offset", "NumValue", "Value", "Description");
                foreach (IELFElem elem in table)
                {
                    if (elem.IsEnumValue())
                    {
                        PrintLineWithEnum(elem);
                    }
                    else
                    {
                        PrintLine(elem);
                    }
                }

                Console.WriteLine();
            }
        }

        public void PrintProgramHeaderTableShort()
        {
            Console.Out.WriteLine();
            Console.Out.WriteLine("ELF Program Header Tables:");
            Console.Out.WriteLine("--------------------------");

            if (!ef.ExistsProgramHeaderTable())
            {
                Console.Out.WriteLine("Program header tables do not exist.");
                return;
            }

            Console.WriteLine(FormatString_ProgTab_Short, "Type", "Offset", "VirtAddr", "PhysAddr", "Filesize", "Memsize", "Flags", "Align");
            foreach (ELFProgramHeaderTable table in ef.GetProgramHeaderTables())
            {
                if (Enum.IsDefined(table.P_Type.GetEnumType(), (int)table.P_Type.Value))
                {
                    Console.WriteLine(FormatString_ProgTab_Short,
                        table.P_Type.GetEnum().ToString(), table.P_Off.ValueString(), table.P_VAddr.ValueString(), table.P_PAddr.ValueString(),
                        table.P_FileSz.ValueString(), table.P_MemSz.ValueString(), table.P_Flags.GetEnumDescription(), table.P_Align.ValueString());
                }
                else
                {
                    Console.WriteLine(FormatString_ProgTab_Short,
                        "Unknown", table.P_Off.ValueString(), table.P_VAddr.ValueString(), table.P_PAddr.ValueString(),
                        table.P_FileSz.ValueString(), table.P_MemSz.ValueString(), table.P_Flags.GetEnumDescription(), table.P_Align.ValueString());
                }
            }
        }

        public void PrintSectionHeaderTableDetailed()
        {
            Console.Out.WriteLine();
            Console.Out.WriteLine("ELF Section Header Tables:");
            Console.Out.WriteLine("--------------------------");

            if (!ef.ExistsSectionHeaderTables())
            {
                Console.Out.WriteLine("Section header tables do not exist.");
                return;
            }

            foreach (ELFSectionHeaderTable table in ef.GetSectionHeaderTables())
            {
                Console.WriteLine("{0,10}", "Section header table " + table.HeaderTableIndex);
                Console.Out.WriteLine("Name: " + table.GetSectionName());
                PrintLine("Name", "Offset", "NumValue", "Value", "Description");
                foreach (IELFElem elem in table)
                {
                    if (elem.IsEnumValue())
                    {
                        PrintLineWithEnum(elem);
                    }
                    else
                    {
                        PrintLine(elem);
                    }
                }
                Console.WriteLine();
            }
        }

        public void PrintSectionHeaderTableShort()
        {
            Console.Out.WriteLine();
            Console.Out.WriteLine("ELF Section Header Tables:");
            Console.Out.WriteLine("--------------------------");

            if (!ef.ExistsSectionHeaderTables())
            {
                Console.Out.WriteLine("Section header tables do not exist.");
                return;
            }

            Console.WriteLine(FormatString_SecTab_Short, "Number", "Name", "Type", "Flags", "Addr", "Offset", "Size", "Link", "Info", "Addr align", "Entry size");
            foreach (ELFSectionHeaderTable table in ef.GetSectionHeaderTables())
            {
                Console.WriteLine(FormatString_SecTab_Short,
                    table.HeaderTableIndex.ToString(), NormalizeLength(table.GetSectionName(), 17), table.Sh_type.GetEnumValue(), table.Sh_flags.GetEnumDescription(),
                    table.Sh_Addr.ValueString(), table.Sh_Offset.ValueString(), table.Sh_Size.ValueString(), table.Sh_Link.ValueString(),
                    table.Sh_Info.ValueString(), table.Sh_Addralign.ValueString(), table.Sh_Entsize.ValueString());
            }
        }

        public void PrintSymbolTablesDetailed()
        {
            Console.Out.WriteLine();
            Console.Out.WriteLine("ELF Program Symbol Tables:");
            Console.Out.WriteLine("--------------------------");

            if (!ef.ExistsSymbolTables())
            {
                Console.Out.WriteLine("Symbol tables do not exist.");
                return;
            }

            foreach (ELFSymbolTable symTable in ef.GetSymbolTables())
            {
                Console.WriteLine("{0,10}", "Symbol section header table " + symTable.SectionHeaderTable.HeaderTableIndex);
                Console.WriteLine("Name: " + symTable.SectionHeaderTable.GetSectionName());
                PrintLine("Name", "Offset", "NumValue", "Value", "Description");

                foreach (ELFSymbolTableEntry symTableEntry in symTable.GetSymbolTableEntries())
                {
                    Console.WriteLine("Index: " + symTableEntry.EntryIndex + " - Name: " + symTableEntry.GetSymbolName());
                    foreach (IELFElem elem in symTableEntry)
                    {
                        if (elem.IsEnumValue())
                        {
                            PrintLineWithEnum(elem);
                        }
                        else
                        {
                            PrintLine(elem);
                        }
                    }
                    Console.WriteLine("Binding: " + symTableEntry.GetBindingInfo() + " - " + symTableEntry.GetBindingInfoAsString());
                    Console.WriteLine("Type: " + symTableEntry.GetTypeInfo() + " - " + symTableEntry.GetTypeInfoAsString());
                    Console.WriteLine("Visibility: " + symTableEntry.GetVisibilityInfo() + " - " + symTableEntry.GetVisibilityInfoAsString());
                    Console.WriteLine();
                }
            }
        }

        public void PrintSymbolTablesShort()
        {
            Console.Out.WriteLine();
            Console.Out.WriteLine("ELF Program Symbol Tables:");
            Console.Out.WriteLine("--------------------------");

            if (!ef.ExistsSymbolTables())
            {
                Console.Out.WriteLine("Symbol tables do not exist.");
                return;
            }

            foreach (ELFSymbolTable symTable in ef.GetSymbolTables())
            {
                Console.WriteLine("{0,10}", "Symbol section header table " + symTable.SectionHeaderTable.HeaderTableIndex);
                Console.WriteLine(FormatString_SymTab_Short, "Number", "Value", "Size", "Type", "Binding", "Visibility", "Index", "Name");
                foreach (ELFSymbolTableEntry symTableEntry in symTable.GetSymbolTableEntries())
                {
                    Console.WriteLine(FormatString_SymTab_Short, symTableEntry.EntryIndex, symTableEntry.St_value.ValueString(), symTableEntry.St_size.ValueString(),
                                symTableEntry.GetTypeInfoAsString(), symTableEntry.GetBindingInfoAsString(), symTableEntry.GetVisibilityInfoAsString(),
                                symTableEntry.St_shndx.ValueString(), symTableEntry.GetSymbolName());
                }
            }
        }

        public void PrintNoteTables()
        {
            Console.Out.WriteLine();
            Console.Out.WriteLine("ELF Note Tables:");
            Console.Out.WriteLine("----------------");

            bool found = false;
            foreach (ELFSectionHeaderTable secTable in ef.GetSectionHeaderTables())
            {
                if (secTable.IsNoteSection())
                {
                    Console.Out.WriteLine();
                    Console.Out.WriteLine("Note table found: " + secTable.GetSectionName());

                    foreach (ELFNote elfNote in secTable.NoteSection)
                    {
                        Console.Out.WriteLine("Name: " + elfNote.Name);
                        Console.Out.WriteLine("Type: " + elfNote.Type);
                        Console.Out.WriteLine("Description: " + elfNote.Description.ValueString());
                    }

                    found = true;
                }
            }

            if (!found)
            {
                Console.Out.WriteLine("Note tables do not exist.");
                return;
            }
        }

        private static void PrintLine(string name, string offsetval, string strValue, string enumVal, string enumDesc)
        {
            Console.WriteLine(FormatString_Default, name, offsetval, strValue, enumVal, enumDesc);
        }

        private static void PrintLineShort(IELFElem elfElem)
        {
            Console.WriteLine(FormatString_Default_Short, elfElem.Name, elfElem.ValueString());
        }

        private static void PrintLineWithEnumShort(IELFElem elfElem)
        {
            Console.WriteLine(FormatString_Default_Short, elfElem.Name, elfElem.GetEnum().ToString());
        }

        private static void PrintLine(IELFElem elfElem)
        {
            if (elfElem.GetType() == typeof(ELFElemByteArray))
            {
                PrintLineByteArray(elfElem);
            }
            else
            {
                PrintLine(elfElem.Name, elfElem.OffsetString, elfElem.ValueString(), "-", "-");
            }
        }

        private static void PrintLineByteArray(IELFElem elfElem)
        {
            Console.WriteLine(FormatString_ByteArray + elfElem.ValueString(), elfElem.Name, elfElem.OffsetString);
        }

        private static void PrintLineWithEnum(IELFElem elfElem)
        {
            if (elfElem.HasEnumFlagsAttibrute())
            {
                PrintLine(elfElem.Name, elfElem.OffsetString, elfElem.ValueString(), elfElem.GetEnumDescription(), "");
            }
            else
            {
                PrintLine(elfElem.Name, elfElem.OffsetString, elfElem.ValueString(), elfElem.GetEnum().ToString(), elfElem.GetEnumDescription());
            }
        }

        private static string NormalizeLength(string value, int maxLength)
        {
            return value.Substring(0, Math.Min(value.Length, maxLength));
        } 
    }
}
