﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ELFViewerConsole.ELFModel;
using ElfViewerConsole.ELFModel.ELFFormat.Enums;

namespace ElfViewerConsole.ELFModel.ELFFormat
{
    class ELFFile : BinaryFile
    {
        public ELFHeader ElfFileHeader
        {
            get;
            private set;
        }

        public bool Valid { set; get; }

        internal ELFProgramHeaderTables ProgramHeaderTables;
        internal ELFSectionHeaderTables SectionHeaderTables;
        internal List<ELFSymbolTable> SymbolTables;



        public ELFFile(string filename) : base(filename) { }

        public override void LoadFile()
        {
            base.LoadFile();
            Valid = true;

            ElfFileHeader = new ELFHeader(this);
            ElfFileHeader.Read(FileData);

            if (Valid)
            {
                ProgramHeaderTables = new ELFProgramHeaderTables(this);
                ProgramHeaderTables.Read();

                SectionHeaderTables = new ELFSectionHeaderTables(this);
                SectionHeaderTables.Read();

                ProcessSymbolTables();
                ProcessNoteTables();
            }
        }

        public bool Is32Bit()
        {
            return ((EELFIdentClass)ElfFileHeader.E_ident_class.GetEnum() == EELFIdentClass.CLASS32);
        }

        public bool Is64Bit()
        {
            return ((EELFIdentClass)ElfFileHeader.E_ident_class.GetEnum() == EELFIdentClass.CLASS64);
        }

        public bool ExistsProgramHeaderTable()
        {
            return ProgramHeaderTables != null && ProgramHeaderTables.Exists;
        }

        public bool ExistsSectionHeaderTables()
        {
            return SectionHeaderTables != null && SectionHeaderTables.Exists;
        }
        

        public IEnumerable<ELFProgramHeaderTable> GetProgramHeaderTables()
        {
            foreach (ELFProgramHeaderTable headerTableEntry in this.ProgramHeaderTables)
            {
                yield return headerTableEntry;
            }
        }

        public IEnumerable<ELFSectionHeaderTable> GetSectionHeaderTables()
        {
            if (ExistsSectionHeaderTables())
            {
                foreach (ELFSectionHeaderTable headerTableEntry in this.SectionHeaderTables)
                {
                    yield return headerTableEntry;
                }
            }
        }

        public IEnumerable<ELFSymbolTable> GetSymbolTables()
        {
            if (ExistsSymbolTables())
            {
                foreach (ELFSymbolTable symTable in this.SymbolTables)
                {
                    yield return symTable;
                }
            }
        }

        public int GetNumOfSections()
        {
            if (ElfFileHeader != null)
            {
                return ElfFileHeader.E_ShNum.Value;
            }
            else
            {
                return 0;
            }
        }

        public bool ExistsSectionNameStringTable()
        {
            return (ElfFileHeader.E_ShStrndx.Value != 0);
        }

        public int GetSectionNameStringTableIndex()
        {
            return ElfFileHeader.E_ShStrndx.Value;
        }

        public bool ExistsSymbolTables()
        {
            return SymbolTables != null && SymbolTables.Count > 0;
        }

        private void ProcessSymbolTables()
        {
            SymbolTables = new List<ELFSymbolTable>();

            foreach (ELFSectionHeaderTable headerTableEntry in this.SectionHeaderTables)
            {
                if (headerTableEntry.Sh_type.GetEnumValue() == EELFShType.SHT_SYMTAB || headerTableEntry.Sh_type.GetEnumValue() == EELFShType.SHT_DYNSYM)
                {
                    /* section header belongs to a symbol table */
                    ELFSymbolTable symTable = new ELFSymbolTable(this, headerTableEntry);
                    symTable.Read();
                    SymbolTables.Add(symTable);
                }
            }
        }

        private void ProcessNoteTables()
        {
            foreach (ELFSectionHeaderTable headerTableEntry in this.SectionHeaderTables)
            {
                if (headerTableEntry.Sh_type.GetEnumValue() == EELFShType.SHT_NOTE)
                {
                    ELFNoteTable ent = new ELFNoteTable(this, headerTableEntry);
                    ent.Read();
                    headerTableEntry.NoteSection = ent;
                }
            }
        }
    }
}
