﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace ELFViewerConsole.ELFModel
{
    class FileDataHandler : IFileAccess
    {
        /************************************************************
         * Variables
         ************************************************************/
        private byte[] fileRawData;

        private uint internalFilePos;

        private IFileAccessEndianess endianType = IFileAccessEndianess.LITTLE_ENDIAN;

        /************************************************************
         * Properties
         ************************************************************/
        public long Length
        {
            get { return fileRawData.Length; }
        }

        public uint Position
        {
            get { return internalFilePos; }
            set { internalFilePos = value; }
        }

        public byte[] RawData
        {
            get { return fileRawData; }
        }

        /************************************************************
         * Methods
         ************************************************************/
        public FileDataHandler(BinaryReader br, long fileLength)
        {
            if (br == null)
            {
                throw new ArgumentNullException();
            }

            if (fileLength > uint.MaxValue)
            {
                throw new FileDataAccessException("Filesize not yet supported.");
            }

            fileRawData = br.ReadBytes((Int32)fileLength);
        }

        public byte ReadByte(uint pos)
        {
            if (pos < 0 || pos >= Length)
            {
                throw new FileDataAccessException();
            }

            return (byte)fileRawData[pos];
        }

        public byte ReadByte()
        {
            return ReadByte(internalFilePos++);
        }

        public ushort ReadUInt16(uint pos)
        {
            if (pos < 0 || pos >= Length - 1)
            {
                throw new FileDataAccessException();
            }

            if (this.endianType == IFileAccessEndianess.LITTLE_ENDIAN)
            {
                return (ushort)(fileRawData[pos] | fileRawData[pos + 1] << 8);
            }
            else
            {
                return (ushort)(fileRawData[pos + 1] | fileRawData[pos] << 8);
            }
        }

        public ushort ReadUInt16()
        {
            ushort result = ReadUInt16(internalFilePos);
            internalFilePos += 2;
            return result;
        }

        public short ReadInt16(uint pos)
        {
            if (pos < 0 || pos >= Length - 1)
            {
                throw new FileDataAccessException();
            }

            if (this.endianType == IFileAccessEndianess.LITTLE_ENDIAN)
            {
                return (short)(fileRawData[pos] | fileRawData[pos + 1] << 8);
            }
            else
            {
                return (short)(fileRawData[pos + 1] | fileRawData[pos] << 8);
            }
        }

        public short ReadInt16()
        {
            short result = ReadInt16(internalFilePos);
            internalFilePos += 2;
            return result;
        }

        public uint ReadUInt32(uint pos)
        {
            if (pos < 0 || pos >= Length - 3)
            {
                throw new FileDataAccessException("Offset " + pos + " is out of file range.");
            }

            if (this.endianType == IFileAccessEndianess.LITTLE_ENDIAN)
            {
                return (uint)(fileRawData[pos] | fileRawData[pos + 1] << 8 |
                              fileRawData[pos + 2] << 16 | fileRawData[pos + 3] << 24);
            }
            else
            {
                return (uint)(fileRawData[pos + 3] | fileRawData[pos + 2] << 8 |
                              fileRawData[pos + 1] << 16 | fileRawData[pos] << 24);
            }
        }

        public uint ReadUInt32()
        {
            uint result = ReadUInt32(internalFilePos);
            internalFilePos += 4;
            return result;
        }

        public int ReadInt32(uint pos)
        {
            if (pos < 0 || pos >= Length - 3)
            {
                throw new FileDataAccessException();
            }

            if (this.endianType == IFileAccessEndianess.LITTLE_ENDIAN)
            {
                return (int)(fileRawData[pos] | fileRawData[pos + 1] << 8 |
                             fileRawData[pos + 2] << 16 | fileRawData[pos + 3] << 24);
            }
            else
            {
                return (int)(fileRawData[pos + 3] | fileRawData[pos + 2] << 8 |
                             fileRawData[pos + 1] << 16 | fileRawData[pos] << 24);
            }
        }

        public int ReadInt32()
        {
            int result = ReadInt32(internalFilePos);
            internalFilePos += 4;
            return result;
        }

        public ulong ReadUInt64(uint pos)
        {
            if (pos < 0 || pos >= Length - 7)
            {
                throw new FileDataAccessException();
            }

            if (this.endianType == IFileAccessEndianess.LITTLE_ENDIAN)
            {
                uint u1 = (uint)(fileRawData[pos] | fileRawData[pos + 1] << 8 |
                                fileRawData[pos + 2] << 16 | fileRawData[pos + 3] << 24);
                uint u2 = (uint)(fileRawData[pos + 4] | fileRawData[pos + 5] << 8 |
                                fileRawData[pos + 6] << 16 | fileRawData[pos + 7] << 24);
                return (ulong)(u1 | (ulong)u2 << 32);
            }
            else
            {
                uint u1 = (uint)(fileRawData[pos + 3] | fileRawData[pos + 2] << 8 |
                                fileRawData[pos + 1] << 16 | fileRawData[pos] << 24);
                uint u2 = (uint)(fileRawData[pos + 7] | fileRawData[pos + 6] << 8 |
                                fileRawData[pos + 5] << 16 | fileRawData[pos + 4] << 24);
                return (ulong)(u2 | (ulong)u1 << 32);
            }
        }

        public ulong ReadUInt64()
        {
            ulong result = ReadUInt64(internalFilePos);
            internalFilePos += 8;
            return result;
        }

        public long ReadInt64(uint pos)
        {
            if (pos < 0 || pos >= Length - 7)
            {
                throw new FileDataAccessException();
            }

            if (this.endianType == IFileAccessEndianess.LITTLE_ENDIAN)
            {
                uint u1 = (uint)(fileRawData[pos] | fileRawData[pos + 1] << 8 |
                                fileRawData[pos + 2] << 16 | fileRawData[pos + 3] << 24);
                uint u2 = (uint)(fileRawData[pos + 4] | fileRawData[pos + 5] << 8 |
                                fileRawData[pos + 6] << 16 | fileRawData[pos + 7] << 24);
                return (long)(u1 | (ulong)u2 << 32);
            }
            else
            {
                uint u1 = (uint)(fileRawData[pos + 3] | fileRawData[pos + 2] << 8 |
                                fileRawData[pos + 1] << 16 | fileRawData[pos] << 24);
                uint u2 = (uint)(fileRawData[pos + 7] | fileRawData[pos + 6] << 8 |
                                fileRawData[pos + 5] << 16 | fileRawData[pos + 4] << 24);
                return (long)(u2 | (ulong)u1 << 32);
            }
        }

        public long ReadInt64()
        {
            long result = ReadInt64(internalFilePos);
            internalFilePos += 8;
            return result;
        }

        public byte[] ReadByteArray(uint pos, uint length)
        {
            if (pos < 0 || pos >= Length - length)
            {
                throw new FileDataAccessException();
            }

            byte[] data = new byte[length];

            for (uint i = pos; i < pos + length; i++)
            {
                data[i - pos] = fileRawData[i];
            }

            return data;
        }

        public byte[] ReadByteArray(uint length)
        {
            byte[] result = ReadByteArray(internalFilePos, length);
            internalFilePos += length;
            return result;
        }

        public void SetEndianess(IFileAccessEndianess type)
        {
            this.endianType = type;
        }

        public string ReadAscii(uint pos, int maxLength)
        {
            int length = (int)maxLength;
            for (int i = 0; i < maxLength; i++)
            {
                if (pos + i >= this.Length || fileRawData[pos + i] == 0)
                {
                    length = i;
                    break;
                }
            }

            StringBuilder sb = new StringBuilder(length + 1);
            for (int i = 0; i < length; i++)
            {
                sb.Append((char)fileRawData[pos + i]);
            }
            return sb.ToString();
        }
    }

    public class FileDataAccessException : IOException
    {
        public FileDataAccessException() : base() { }
        public FileDataAccessException(string message) : base(message) { }
    }
}
